/*
===========================================================================
Copyright (C) 2000 - 2013, Raven Software, Inc.
Copyright (C) 2001 - 2013, Activision, Inc.
Copyright (C) 2013 - 2015, OpenJK contributors

This file is part of the OpenJK source code.

OpenJK is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, see <http://www.gnu.org/licenses/>.
===========================================================================
*/

#ifndef __strip_objectives_h
#define __strip_objectives_h


/********************************************************************************

the objectives... duh...

********************************************************************************/


#define PACKAGE_OBJECTIVES		0x08

enum
{
	OBJECTIVES_KEJIM_POST_OBJ1 = 0x0800,
	OBJECTIVES_KEJIM_POST_OBJ2 = 0x0801,
	OBJECTIVES_KEJIM_BASE_OBJ1 = 0x0802,
	OBJECTIVES_KEJIM_BASE_OBJ2 = 0x0803,
	OBJECTIVES_KEJIM_BASE_OBJ3 = 0x0804,
	OBJECTIVES_ARTUS_MINE_OBJ1 = 0x0805,
	OBJECTIVES_ARTUS_MINE_OBJ2 = 0x0806,
	OBJECTIVES_ARTUS_MINE_OBJ3 = 0x0807,
	OBJECTIVES_ARTUS_DETENTION_OBJ1 = 0x0808,
	OBJECTIVES_ARTUS_DETENTION_OBJ2 = 0x0809,
	OBJECTIVES_ARTUS_TOPSIDE_OBJ1 = 0x080a,
	OBJECTIVES_ARTUS_TOPSIDE_OBJ2 = 0x080b,
	OBJECTIVES_YAVIN_TEMPLE_OBJ1 = 0x080c,
	OBJECTIVES_YAVIN_TRIAL_OBJ1 = 0x080d,
	OBJECTIVES_YAVIN_TRIAL_OBJ2 = 0x080e,
	OBJECTIVES_NS_STREETS_OBJ1 = 0x080f,
	OBJECTIVES_NS_STREETS_OBJ2 = 0x0810,
	OBJECTIVES_NS_STREETS_OBJ3 = 0x0811,
	OBJECTIVES_NS_HIDEOUT_OBJ1 = 0x0812,
	OBJECTIVES_NS_HIDEOUT_OBJ2 = 0x0813,
	OBJECTIVES_NS_STARPAD_OBJ1 = 0x0814,
	OBJECTIVES_NS_STARPAD_OBJ2 = 0x0815,
	OBJECTIVES_NS_STARPAD_OBJ3 = 0x0816,
	OBJECTIVES_NS_STARPAD_OBJ4 = 0x0817,
	OBJECTIVES_NS_STARPAD_OBJ5 = 0x0818,
	OBJECTIVES_BESPIN_UNDERCITY_OBJ1 = 0x0819,
	OBJECTIVES_BESPIN_UNDERCITY_OBJ2 = 0x081a,
	OBJECTIVES_BESPIN_STREETS_OBJ1 = 0x081b,
	OBJECTIVES_BESPIN_STREETS_OBJ2 = 0x081c,
	OBJECTIVES_BESPIN_PLATFORM_OBJ1 = 0x081d,
	OBJECTIVES_BESPIN_PLATFORM_OBJ2 = 0x081e,
	OBJECTIVES_CAIRN_BAY_OBJ1 = 0x081f,
	OBJECTIVES_CAIRN_BAY_OBJ2 = 0x0820,
	OBJECTIVES_CAIRN_ASSEMBLY_OBJ1 = 0x0821,
	OBJECTIVES_CAIRN_REACTOR_OBJ1 = 0x0822,
	OBJECTIVES_CAIRN_REACTOR_OBJ2 = 0x0823,
	OBJECTIVES_CAIRN_DOCK1_OBJ1 = 0x0824,
	OBJECTIVES_CAIRN_DOCK1_OBJ2 = 0x0825,
	OBJECTIVES_DOOM_COMM_OBJ1 = 0x0826,
	OBJECTIVES_DOOM_COMM_OBJ2 = 0x0827,
	OBJECTIVES_DOOM_COMM_OBJ3 = 0x0828,
	OBJECTIVES_DOOM_DETENTION_OBJ1 = 0x0829,
	OBJECTIVES_DOOM_DETENTION_OBJ2 = 0x082a,
	OBJECTIVES_DOOM_SHIELDS_OBJ1 = 0x082b,
	OBJECTIVES_DOOM_SHIELDS_OBJ2 = 0x082c,
	OBJECTIVES_YAVIN_SWAMP_OBJ1 = 0x082d,
	OBJECTIVES_YAVIN_SWAMP_OBJ2 = 0x082e,
	OBJECTIVES_YAVIN_CANYON_OBJ1 = 0x082f,
	OBJECTIVES_YAVIN_CANYON_OBJ2 = 0x0830,
	OBJECTIVES_YAVIN_COURTYARD_OBJ1 = 0x0831,
	OBJECTIVES_YAVIN_COURTYARD_OBJ2 = 0x0832,
	OBJECTIVES_YAVIN_FINAL_OBJ1 = 0x0833,
	OBJECTIVES_KEJIM_POST_OBJ3 = 0x0834,
	OBJECTIVES_KEJIM_POST_OBJ4 = 0x0835,
	OBJECTIVES_KEJIM_POST_OBJ5 = 0x0836,
	OBJECTIVES_ARTUS_DETENTION_OBJ3 = 0x0837,
	OBJECTIVES_DOOM_COMM_OBJ4 = 0x0838,
	OBJECTIVES_DOOM_SHIELDS_OBJ3 = 0x0839,
	OBJECTIVES_DEMO_OBJ1 = 0x083a,
	OBJECTIVES_DEMO_OBJ2 = 0x083b,
	OBJECTIVES_DEMO_OBJ3 = 0x083c,
	OBJECTIVES_DEMO_OBJ4 = 0x083d
};


#endif // __strip_objectives_h

