// ICARUS Engine Interface File
//
//	This file is the only section of the ICARUS systems that
//	is not directly portable from engine to engine.
//
//	-- jweier

#include "Interface.h"

void Interface_Init( interface_export_t *pe )
{

	//TODO: This is where you link up all your functions to the engine

	//Example:
	//
	// pe->I_GetEntityByName = ENGINE_GetEntityByName;
}
